#! /usr/bin/perl -w
$| = 1;

##
##  Install UCS on your system, i.e. adjust Perl binary to use and library paths.
##  Usage: perl Install.perl
##

## Filename: Install.perl
## Modified: Sun Jan 25 20:29:07 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: install UCS by adjusting paths in Perl scripts

use FileHandle;
use FindBin;
use Config;

$Force = 0;					# install even when prerequisites are not satisfied
$Force = 1
  if @ARGV == 1 and $ARGV[0] =~ /^--?f(orce)?$/i;

# check Perl version (5.6.1 required)
if (not defined eval 'require 5.006_001; 1') {
  chomp($msg = $@);
  $msg =~ s/,.*$//;		# cut off "explanation" after comma
  if ($Force) {
    warn "Perl version error: $msg. Installing anyway.\n";
  }
  else {
    die "Perl version error: $msg. Aborted.\n";
  }
}

# check for required modules
@mods = qw<Expect>;		
@missing = grep { not defined eval "use $_; 1" } @mods;
if (@missing > 0) {
  if ($Force) {
    warn "Warning: one or more modules required by UCS are not installed on your system.\n",
      "\t[missing module(s): @missing]\n",
	"Installing anyway (--force).  USE AT YOUR OWN RISK!\n";
  }
  else {
    die "Error: one or more modules required by UCS are not installed on your system.\n",
      "\t[missing module(s): @missing]\n";
  }
}

# check for optional modules (ask user whether to proceed)
@mods = qw<Tk Tk::Pod Term::ReadKey>;		
@missing = grep { not defined eval "use $_; 1"} @mods;
if (@missing > 0) {
  print "Warning: one or more recommended modules are not installed on your system.\n",
    "UCS/Perl works without these modules, but some functions may not be available.\n",
      "\t[missing module(s): @missing]\n";
  if (not $Force) {
    yes_or_no("Continue without these modules?")
      or die "Installation aborted.\n";
  }
}

# check for expected future dependencies (just issue a warning)
@mods = qw<Tk::TableMatrix>;
# @mods = qw<Math::CDF PDL::GSLSF::GAMMA>;  # not sure whether I'm going to use these at all
@missing = grep { not defined eval "use $_; 1" } @mods;
print
  "Notice: the following modules, which may be required by future releases of\n",
  "UCS/Perl, are not installed on your system. Currently, there is no need for\n",
  "these modules, so I will continue with the installation.\n",
  "\t[missing module(s): @missing]\n"
  if @missing > 0;

# find Perl interpreter (absolute path for #! line)
$Perl = $Config{"perlpath"};
die "Yikes! Your Perl interpreter ($Perl) doesn't seem to be an executable program. Aborted.\n"
  unless -x $Perl;
die "Error: need absolute path to Perl binary ($Perl). Aborted.\n"
  unless $Perl =~ /^\//;

# find R binary in user's path
@path = split /:/, $ENV{PATH};
$Rbin = undef;
foreach $dir (@path) {
  if (-f "$dir/R" and -x "$dir/R") {
    $Rbin = "$dir/R";
    last;
  }
}

if ($Force) {
  if (not defined $Rbin) {
    warn "Warning: I couldn't find the R program in your search path.\n",
      "I will continue with the installation anyway (--force), but the\n",
	"UCS/R part will not be available and UCS/Perl will not function\n",
	  "reliably.  USE AT YOUR OWN RISK!\n";
    $Rbin = "R";
  }
}
else {
  die "Error: Couldn't find R program in your search path. Please make sure\n",
    "that R is installed and insert the appropriate directory in your search\n",
      "path if necessary. (You should be able to run R by entering 'R' on the\n",
	"command line.)\n"
	  unless defined $Rbin;
  die
    "Error: need absoluted path to R binary ($Rbin). Aborted.\n"
  unless $Rbin =~ /^\//;
}

# determine library directory
$Basedir = $FindBin::RealBin;
die "Couldn't determine UCS base directory ($Basedir is wrong). Aborted.\n"
  unless -d "$Basedir/Perl/lib" and -f "$Basedir/Install.perl";
die "Error: need absolute Path to UCS base directory ($Basedir). Aborted.\n"
  unless $Basedir =~ /^\//;
$Libdir = $Basedir."/Perl/lib";
$Bindir = $Basedir."/Perl/bin";
$Tooldir = $Basedir."/Perl/tools";

$Version = join(".", @Config{qw<PERL_REVISION PERL_VERSION PERL_SUBVERSION>});

# print summary of configuration information
print "\n";
print "=== UCS System Configuration ===\n";
print "\n";
print "Perl Version:       $Version\n";
print "Perl Interpreter:   $Perl\n";
print "R binary:           $Rbin\n";
print "UCS Base Directory: $Basedir\n";
print "\n";
print "================================\n";
print "\n";

# clean up Backup files
@Backups = (glob("$Bindir/*~"), glob("$Tooldir/*~"));
$N = @Backups;
if ($N > 0) {
  print "Deleting $N emacs backup file(s).\n";
  foreach $file (@Backups) { unlink $file }
}

# assemble shebang line to insert
$shebang = "#!$Perl -w -I$Libdir";
print "Perl command:       $shebang\n";

# check which files require substitution
@Files = (glob("$Bindir/*"), glob("$Tooldir/*"));
$N = @Files;
print "Substituting #! line in $N files ...\n";

# substitute shebang line in Perl scripts
foreach $file (@Files) {
  substitute($file, qr/^\#!.*perl/, $shebang);
  system "chmod a+rx $file";			# make sure the files are marked as executables
}

# substitute base directory in UCS.pm (UCS/Perl configuration)
print "Editing configuration files ...\n";
$dirline = "\$BaseDir = \"$Basedir\";\t # in-place edit by Install.perl\n";
substitute("$Libdir/UCS.pm", qr/^\s*\$BaseDir\s*=/, $dirline);
$cmdline = "\$PerlCmd = '$shebang';\t # in-place edit by Install.perl\n";
substitute("$Libdir/UCS.pm", qr/^\s*\$PerlCmd\s*=/, $cmdline);
$cmdline = "\$Rbin = '$Rbin';\t # in-place edit by Install.perl\n";
substitute("$Libdir/UCS.pm", qr/^\s*\$Rbin\s*=/, $cmdline);

# substitute base directory in ucs.R (UCS/R configuration)
substitute("$Basedir/R/lib/ucs.R", qw/^\s*ucs\.basedir\s*<-/, 
	   "ucs.basedir <- \"$Basedir\"\t # in-place edit by Install.perl\n");

# now test installation by running the ucs-config command
print "\nTesting installation ...\n";
$result = system "$Bindir/ucs-config";
die "Oops! Something went wrong -- installation aborted. :o(\n"
  if $result != 0;

# that's all we have to do
print "Installation successful.\n\n";



## substitute($file, $regex, $line);
##  in-place edit $file, substituting every line matching $regex by $line;
##  $file is backed up as "$file~"; dies unless exactly one substitution is made
sub substitute {
  my ($file, $rx, $line) = @_;
  $line .= "\n"					# make sure the newline character isn't missing
    unless $line =~ /\n$/;
  print "  [$file]\n";
  my $fh = new FileHandle $file;		# read file into array (list of lines)
  die "Can't read '$file'. Aborted.\n"
    unless defined $fh;
  my @lines = <$fh>;
  die "Read error.\n"
    unless @lines >= 6;			        # rough check :o)   [might check against "wc -l" instead]
  $fh->close;
  my $substitutions = 0;			# should replace exactly 1 line
  foreach my $l (@lines) {
    if ($l =~ $rx) {
      $l = $line;
      $substitutions++;
    }
  }
  die "!! substitution error ($substitutions line(s) modified). Aborted.\n"
    unless $substitutions == 1;
  my $backup = "$file~";			# back up original file
  unlink $backup
    if -f $backup;
  system "cp -p $file $backup";			# we'll ignore failure to make a backup copy
  $fh = new FileHandle "> $file";		# write modified lines back to file
  die "Write error.\n"
    unless defined $fh;
  map { $fh->print($_) } @lines;
  $fh->close
    or die "Write error. '$file' may be damaged!\n";
  system "touch -r $backup $file";		# reset modification time of edited file 
}

## $ok = yes_or_no($prompt);
##   prompt user for yes / no answer
sub yes_or_no {
  my $prompt = shift;
  $prompt = "Are you sure you want to continue ?"
    unless defined $prompt;
  print "$prompt (yes/no) ";
  while (my $answer = <STDIN>) {
    chomp $answer;
    $answer = lc $answer;
    if ("yes" eq $answer) {
      return 1;
    } elsif ("no" eq $answer or "n" eq $answer) {
      return 0;
    } else {
      print "Please answer 'yes' or 'no': ";
    }
  }
}
