#!/usr/bin/perl -w -I/home/evert/UCS/System/Perl/lib
# -*-cperl-*-
## Filename: ucs-print-documentation.perl
## Modified: Sun Jan 25 14:45:59 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: print all UCS/Perl PODs in PostScript format
$| = 1;

use UCS;
use UCS::File;

use File::Find;

use sigtrap qw<die normal-signals>;

die "\nUsage:  ucs-print-documentation.perl  <prefix>\n\n",
  "Converts the manual pages of all UCS/Perl modules and programs\n",
  "into PostScript format for printing.  The manual pages will be\n",
  "written into a series of files named '<prefix>-NNN.ps'.\n\n"
  unless @ARGV == 1;

$Prefix = shift @ARGV;
$Number = 1;

$BinDir = $UCS::PerlDir."/bin";
$LibDir = $UCS::PerlDir."/lib";
$LibDir =~ s/Perl/perl/g;			# need this trick to get manpage headers right

# print general (introductory) PODs
print "Printing general documents: ";
@pods = glob "$LibDir/pod/*.pod";
foreach $pod (@pods) { print_pod($pod) }
print "\n";

# print documentation for UCS programs
print "Printing docs of programs:  ";
@progs = grep {not /~$/} glob "$BinDir/*";
@progs = grep {not /test/i} @progs;		# skip test programs without documentation
foreach $prog (@progs) { print_pod($prog) }
print "\n";

# finally, scan UCS/Perl lib dir for Perl modules and print their PODs
print "Printing docs of modules:   ";
find({
      wanted => sub { print_pod($File::Find::name) if /\.pm$/ }, 
      no_chdir => 1,
     },
     $LibDir);
print "\n";

# That's it.


# print one POD document (output filename will be chosen automatically)
sub print_pod {
  my $name = shift;
  my $outname = sprintf "%s-%03d.ps", $Prefix, $Number;
  UCS::File::ShellCmd("$BinDir/ucsdoc -ps $name > $outname");
  print "[$Number] ";
  $Number++;
}
