##
##  UCS / The R ToolBox
##

## Filename: base.R
## Modified: Mon Dec 15 15:39:34 2003 (evert)   
##   Author: Stefan Evert
##  Purpose: Basic functions and constants for handling UCS data sets



## AM   (data frame of character strings)
## keys <- AM$key
##   list of association measures built into UCS, long names & corresponding data set variables
##   [row format: "key" = internal name, "long" = long name, "var" = variable name, "rvar" = variable name for ranking]
AM.matrix <-
  matrix(c("frequency", "cooccurrence frequency",
           "random", "random scores between 0 and 1",
           "MI", "MLE of pointwise mutual information (MI)",
           "relative.risk", "MLE of log relative risk coefficient",
           "Dice", "MLE of Dice coefficient",
           "gmean", "MLE of geometric mean coefficient",
           "MS", "MLE of minimum sensitivity (MS) coefficient",
           "Jaccard", "MLE of Jaccard coefficient",
           "odds.ratio", "MLE of log odds ratio",
           "odds.ratio.disc", "'discounted' log odds ratio",
           "MI2", "MLE of MI, but with numerator squared",
           "MI3", "MLE of MI, but with numerator cubed",
           "z.score", "z-score for O11 against E11",
           "z.score.pv", "z-score for O11 against E11 (p-value)",
           "z.score.corr", "z-score with Yates' correction",
           "z.score.corr.pv", "z-score with Yates' correction (p-value)",
           "t.score", "t-test for comparison of O11 against E11",
           "t.score.pv", "t-test for comparison of O11 against E11 (p-value)",
           "chi.squared.tt", "Pearson's chi-squared statistic (two-sided)",
           "chi.squared.corr.tt", "Pearson's chi-squared statistic (two-sided, p-value)",
           "chi.squared.tt.pv", "Pearson's chi-squared statistic (two-sided, p-value)",
           "chi.squared.corr.tt.pv", "Pearson's chi-squared statistic (two-sided, p-value)",
           "chi.squared", "Pearson's chi-squared statistic (one-sided)",
           "chi.squared.corr", "Pearson's chi-squared statistic (one-sided, Yates' correction)",
           "chi.squared.pv", "Pearson's chi-squared statistic (one-sided, p-value)",
           "chi.squared.corr.pv", "Pearson's chi-squared statistic (one-sided, Yates' correction, p-value)",
           "log.likelihood.tt", "Dunning's log-likelihood ratio (two-sided likelihood ratio test)",
           "log.likelihood.tt.pv", "Dunning's log-likelihood ratio (two-sided, p-value)",
           "log.likelihood", "Dunning's log-likelihood ratio (one-sided)",
           "log.likelihood.pv", "Dunning's log-likelihood ratio (one-sided, p-value)",
           "multinomial.likelihood.pv", "Multinomial likelihood of the observed contingency table",
           "binomial.likelihood.pv", "Binomial likelihood for null distribution of O11",
           "Poisson.likelihood.R.pv", "Poisson likelihood for null distribution of O11 (R implementation)",
           "Poisson.likelihood.pv", "Poisson likelihood for null distribution of O11 (Perl implementation)",
           "hypergeometric.likelihood.pv", "hypergeometric likelihood of observed table with fixed marginals",
           "Poisson.Stirling", "Poisson likelihood simplified with Stirling's formula",
           "binomial.pv", "binomial test for null distribution of O11 (one-sided, p-value)",
           "Poisson.pv", "Poisson test for null distribution of O11 (one-sided, p-value)",
           "Fisher.pv", "Fisher's exact test (one-sided, p-value)",
           "average.MI", "average mutual information between indicator variables Y and Z",
           "local.MI", "contribution to average mutual information of full cooccurrence data",
           "MI.conf", "conservative estimate of mutual information (confidence interval)",
           "MI.conf.2", "conservative estimate of mutual information (sig.level = 1E-2)",
           "MI.conf.3", "conservative estimate of mutual information (sig.level = 1E-3)",
           "MI.conf.5", "conservative estimate of mutual information (sig.level = 1E-5)",
           "MI.conf.10", "conservative estimate of mutual information (sig.level = 1E-10)",
           "MI.conf.50", "conservative estimate of mutual information (sig.level = 1E-50)",
           "MI.conf.100", "conservative estimate of mutual information (sig.level = 1E-100)",
           "MI.conf.1000", "conservative estimate of mutual information (sig.level = 1E-1000)",
           "Poisson.mu.pv", "Poisson test for null hypothesis mu = x (one-sided, p-value)",
           "Poisson.mu.2.pv", "Poisson test for null hypothesis mu = 2 (one-sided, p-value)",
           "Poisson.mu.3.pv", "Poisson test for null hypothesis mu = 3 (one-sided, p-value)",
           "Poisson.mu.5.pv", "Poisson test for null hypothesis mu = 5 (one-sided, p-value)",
           "Poisson.mu.10.pv", "Poisson test for null hypothesis mu = 10 (one-sided, p-value)",
           "Poisson.mu.50.pv", "Poisson test for null hypothesis mu = 50 (one-sided, p-value)",
           "Poisson.mu.100.pv", "Poisson test for null hypothesis mu = 100 (one-sided, p-value)",
           "Poisson.mu.1000.pv", "Poisson test for null hypothesis mu = 1000 (one-sided, p-value)",
           "Poisson.mu.10000.pv", "Poisson test for null hypothesis mu = 10000 (one-sided, p-value)",
           ),
         ncol=2,
         byrow=TRUE)
AM <- data.frame(key=I(AM.matrix[,1]), long=I(AM.matrix[,2]))
rownames(AM) <- AM$key
AM$var <- I(paste("am", AM$key, sep="."))
AM$rvar <- I(paste("r", AM$key, sep="."))


## if (am.exists("chi.squared")) { ... }
## am.exists("logl2", fail=TRUE)
##   check whether association measure (represented by its key) exists in AM table;
##   aborts with error message if fail=TRUE and AM doesn't exist
am.exists <- function (key, fail=FALSE) {
  ok <- key %in% AM$key
  if (fail && !ok) stop(paste("I don't know the association measure <", key, "> !", sep=""))
  return(ok)
}



## am.var <- am.key2var("chi.squared")
## r.var <- am.key2rvar("chi.squared")
## longname <- am.key2long("chi.squared")
## key <- am.var2key("am.chi.squared")
## keys <- am.keys()
##   retrieve names of data set variables for association scores (am.key2var) and ranking (am.key2rvar),
##   or the long name (am.key2long) of an association measure identified by its key;
##   derive an AM key from the name of the corresponding data set variable (am.var2key);
##   am.keys() returns a list of all registered AM keys (same as AM$key);
am.key2var <- function (key) if (am.exists(key)) AM[key, "var"] else NA
am.key2rvar <- function (key) if (am.exists(key)) AM[key, "rvar"] else NA
am.key2long <- function (key) if (am.exists(key)) AM[key, "long"] else NA
am.var2key <- function (var) { i <- match(var, AM$var); if (is.na(i)) i <- match(var, AM$rvar); if (is.na(i)) NA else AM$key[i] }
am.keys <- function () AM$key



## if (am.in.ds(glaw, "chisq" [, rank=TRUE])) { ... }
## am.in.ds(glaw, "logl", fail=TRUE)
##   check whether data set contains scores for given AM (or ranking if rank=TRUE);
##   aborts with error message if fail=TRUE and AM isn't found in data set; always aborts if AM doesn't exist
am.in.ds <- function (ds, key, rank=FALSE, fail=FALSE) {
  am.exists(key, fail=TRUE)             # stop with error message if AM doesn't exist
  var <- (if (rank) am.key2rvar(key) else am.key2var(key))
  ok  <- var %in% names(ds)
  if (fail && !ok) {
    stop(paste("Data set doesn't contain ",(if (rank) "ranking" else "scores")," for AM <",key,">", sep=""))
  }
  return(ok)
}


## add.am(key="chi.squared.ind", long="Pearson's chi-squared test of independence")
##   add AM to registered measures (both parameters are mandatory)
add.am  <-  function (key, long) {
  if (missing(key) || missing(long)) {
    stop('Usage: add.am(key="chi.squared.ind", long="Pearson\'s chi-squared test of independence")')
  }
  if (am.exists(key)) stop (paste("Cannot redefine AM <",key,"> !", sep=""))
  AM <<- rbind(AM, c(key, long, paste("am.",key,sep=""), paste("r.",key,sep="")))
  rownames(AM)[nrow(AM)] <<- key
}



## glaw <- read.ds.gz("DataSet/glaw.ds.gz")
##   read (gzipped) data set (.ds or .ds.gz) into data frame;
##   compressed data set files are automatically decompressed, and comment lines are ignored;
##   the variables l1 and l2 are converted to character vectors, and b.* to logicals (f.* remain factors)
read.ds.gz <- function (filename) {
  tmp <- tempfile("UCS-R-TEMP.")
  on.exit(unlink(tmp))
  if (length(grep("\\.gz$", filename))) {
    ## uncompress gzipped data set to temporary file and remove comments
    system(paste("gzip -dc", filename, "| grep -v '^#' >", tmp))
  } else {
    ## copy uncompressed data set to temporary file and remove comments
    system(paste("grep -v '^#'", filename, ">", tmp))
  }
  ## read temporary file into R data frame
  df <- read.delim(tmp)
  ## 'magic' data type conversions
  df$l1 <- as.character(df$l1)
  df$l2 <- as.character(df$l2)
  booleans <- grep("^b\\.", names(df)) # Boolean attributes in the data frame
  for (att in booleans) {
    df[att] <- as.logical(df[[att]])
  }
  ## return the data frame (invisible)
  invisible(df)
}



## ordering <- order.by.am(glaw, "chi.squared" [, randomise=FALSE])
##   order data set by given AM (randomly breaking ties if randomise==TRUE (default),
##   using the variable am.random, which must be present);  returns row index vector
order.by.am <- function (ds, am, randomise=TRUE) {
  am.in.ds(ds, am, fail=TRUE)           # check that data set contains requested AM
  am.var <- am.key2var(am)              # get data set variable name for AM
  if (randomise) {
    am.in.ds(ds, "random", fail=TRUE) 
    rand.var <- am.key2var("random")    # use 'am.random' variable for breaking ties
    order(-ds[,am.var], ds[,rand.var])  # sort into descending order 
  }
  else {
    order(-ds[,am.var], ds[,"id"])      # sort rows with identical AM scores in original (i.e. ID) order
  }
}



## glaw <- add.ranks(glaw [, keys=c("MI", "MI2")] [, randomise=TRUE] [, overwrite=TRUE])
##   add rankings for given association measures (identified by their keys) to data set;
##   if the parameters keys is omitted, rankings will be computed for all registered AMs;
##   measures whose scores are not annotated in the data set are silently ignored,
##   and existing rankings will be overwritten; if randomise==TRUE (default), breaks
##   ties in the ranking randomly (using the variable am.random, which must be present);
##   if overwrite==FALSE, existing rankings won't be recomputed, which is useful when
##   adding association scores at a later time (computing them within R)
add.ranks <- function (ds, keys=am.keys(), randomise=TRUE, overwrite=TRUE) {
  n.rows <- nrow(ds)
  for (am in keys) {
    if (am.in.ds(ds, am)) {             # skip AMs not defined in data set
      if (overwrite || !am.in.ds(ds, am, rank=TRUE)) {
        cat(am, " ", sep="")
        ordering <- order.by.am(ds, am, randomise)      # list of row indices sorted by AM scores
        ranks <- vector(mode="numeric", length=n.rows)  # create dummy vector of appropriate length
        ranks[ordering] <- 1:n.rows       # this assigns the correct rank to each row, I believe
        am.rvar <- am.key2rvar(am)        # UCS variable name for ranking
        ds[[am.rvar]] <- ranks
      }
    }
  }
  cat("\n")
  invisible(ds)                         # don't forget to return modified data set :o)
}



## pr.tbl <- precision.recall(glaw, "chi.squared" [, step=1, first=1, cut=<none>, window=100])
## compute precision/recall table for specified AM; the data set must contain a
## ranking for the given AM and a user-defined variable "b.TP" marking TPs;
## optionally, increase size of n-best lists by <step> and skip lists with n < <first>;
## if <cut> is specified, compute n-best lists up to n=<cut> only, but pretend this
## corresponds to 100% of the candidate list (used for combined data from multiple n-best lists);
## in addition, local precision is computed, averaging over <window> points to the
## left and right of the current value of n; local precision values for n <= <window>
## and for n > max(n) - <window> are undefined
##   returns data frame with columns
##     n         =  number of pairs in n-best list
##     perc      =  as percentage of full data set
##     tp        =  number of TPs in n-best list
##     precision =  precision value for this n-best list
##     recall    =  recall value for this n-best list (relative recall if <cut> was specified)
##     lp        =  local precision, averaged over symmetric window of specified size
precision.recall <- function (ds, am, step=1, first=1, cut=NULL, window=100) {
  am.in.ds(ds, am, fail=TRUE)           # check whether requested AM exists in data set
  if (! am.in.ds(ds, am, rank=TRUE)) {  # ... and try to be helpful about ranks
    stop("You must add ranks to your data set first! (with the add.ranks() function)")
  }
  if (!("b.TP" %in% names(ds))) {       # be helpful about defining TPs, too :o)
    stop('You must define TPs for your data set first! (variable named "b.TP")')
  }
  am.rvar <- am.key2rvar(am)
  order <- order(ds[[am.rvar]])         # re-order data set by AM's ranking
  if (! missing(cut)) {
    order <- order[1:cut]               # take only first <cut> elements of sig. list
  }
  tp <- ds[order, "b.TP"]               # (Boolean) ordered list of FPs and TPs
  size <- length(tp)                    # size of data set
  n <- 1:size                           # number of pair types in n-best list
  perc <- 100 * n / size                # corresponding percentage of full data set
  tp.count <- cumsum(tp)                # number of TPs in n-best list
  tp.total <- sum(tp)                   # total number of TPs in sample
  prec <- 100 * tp.count / n            # precision values (percentage)
  recall <- 100 * tp.count / tp.total   # recall values (percentage)
  if (2 * window < size) {
    tp.plus <- c(tp, rep(0, 2 * window + 1))  # tp.plus[n + window] = TP enters window around n 
    tp.minus <- c(rep(0, 2 * window + 1), tp) # tp.minus[n + window] = TP leaves window around n
    lp <- 100 * (cumsum(tp.plus) - cumsum(tp.minus)) / (2 * window + 1)
    ## select lp values corresponding to n = window+1 .. size-window and pad with 0's
    lp <- c(rep(0, window), lp[(2 * window + 1) : size], rep(0, window)) 
  }
  else {
    lp <- rep(0, size)
  }
  pr.rec <- data.frame(n=n,             # combine results into data frame
                       perc=perc, tp=tp.count, precision=prec, recall=recall, lp=lp)
  if ((step > 1) || (first > 1)) {
    subset <- rev(seq(size, first, by=-step)) # construct index seq. from <first> exactly up to <size> by <step>
    invisible(pr.rec[subset,])
  } else {
    invisible(pr.rec)                     # don't print automatically, of course
  }
}
